"""
These are not callbacks, but if you look for callback most likely you look for these
"""

import io
from pathlib import Path

import numpy as np
import torch


class EarlyStopping():

    def __init__(self, patience=10, delta=0.0001):

        self.patience = patience
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.delta = delta


    def __call__(self, val_loss):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.counter = 0

    def we_should_stop(self):
        return self.early_stop


class Checkpoint:
    def __init__(self, dirname: Path = None, id: str = None, in_memory: bool = True, save_best: bool = True):
        self.dirname = dirname
        self.id = id
        self.curr_best_loss = np.inf
        self.save_best = save_best
        self.in_memory = in_memory
        if not self.in_memory:
            assert self.dirname is not None
            assert self.id is not None
            self.path = Path(self.dirname) / f"params_{self.id}.pt"
            self.path.parent.mkdir(exist_ok=True)
        else:
            self.path = None

        self.buffer = io.BytesIO()
        self.best_model = None
        self.best_epoch = None
    
    def reset(self):
        self.curr_best_loss = np.inf
        self.best_model = None
        self.best_epoch = None
        self.buffer = io.BytesIO()

    def __call__(self, net, loss: float, epoch: int):
        if loss < self.curr_best_loss:
            self.curr_best_loss = loss
            self.best_model = net.state_dict()
            self.best_epoch = epoch
            if self.save_best:
                self.save()
    
    def save(self):
        if self.in_memory:
            self.buffer = io.BytesIO()
            torch.save(self.best_model, self.buffer)  # nosec B614
        else:
            torch.save(self.best_model, self.path)  # nosec B614

    def load(self):
        if self.in_memory:
            # Reset the buffer's position to the beginning
            self.buffer.seek(0)
            return torch.load(self.buffer, weights_only=True)  # nosec B614
        else:
            return torch.load(self.path)  # nosec B614




class EpochStatistics():

    def __init__(self) -> None:
        self.n = 0
        self.loss = 0
        self.score = 0
        
    def update(self, loss, score, n):
        self.n += n
        self.loss += loss * n
        self.score += score * n

    def get(self):
        return self.loss / self.n, self.score / self.n
    


class TrackOutput():

    def __init__(self) -> None:
        self.y_true: list[np.ndarray] = []
        self.y_pred: list[np.ndarray] = []

    def update(self, y_true: np.ndarray, y_pred: np.ndarray):
        self.y_true.append(y_true)
        self.y_pred.append(y_pred)

    def get(self):
        return np.concatenate(self.y_true, axis=0), np.concatenate(self.y_pred, axis=0)


from enum import IntEnum

try:
    from enum import StrEnum
except ImportError:
    # StrEnum is not available in Python < 3.11, so we create a compatible version
    from enum import Enum
    class StrEnum(str, Enum):
        """
        Enum where members are also (and must be) strings
        """
        def __new__(cls, value):
            if not isinstance(value, str):
                raise TypeError(f"{value!r} is not a string")
            return super().__new__(cls, value)

        def __str__(self):
            return self.value


class Task(StrEnum):
    CLASSIFICATION = "classification"
    REGRESSION = "regression"


class FeatureType(StrEnum):
    NUMERICAL = "numerical"
    CATEGORICAL = "categorical"
    MIXED = "mixed"


class SearchType(StrEnum):
    DEFAULT = "default"
    RANDOM = "random"


class DatasetSize(IntEnum):
    SMALL = 1000
    MEDIUM = 10000
    LARGE = 50000


class DataSplit(StrEnum):
    TRAIN = "train"
    VALID = "valid"
    TEST = "test"


class Phase(StrEnum):
    TRAINING = "training"
    VALIDATION = "validation"
    TESTING = "testing"


class ModelName(StrEnum):
    PLACEHOLDER = "_placeholder_"   # This is a placeholder for the current running model
    FT_TRANSFORMER = "FT-Transformer"
    TABPFN = "TabPFN"
    FOUNDATION = "Foundation"
    FOUNDATION_FLASH = "FoundationFlash"
    TAB2D = "Tab2D"
    TAB2D_COL_ROW = "Tab2D_COL_ROW"
    TAB2D_SDPA = "Tab2D_SDPA"
    SAINT = "SAINT"
    MLP = "MLP"
    MLP_RTDL = "MLP-rtdl"
    RESNET = "Resnet"
    RANDOM_FOREST = "RandomForest"
    XGBOOST = "XGBoost"
    CATBOOST = "CatBoost"
    LIGHTGBM = "LightGBM"
    GRADIENT_BOOSTING_TREE = "GradientBoostingTree"
    HIST_GRADIENT_BOOSTING_TREE = "HistGradientBoostingTree"
    LOGISTIC_REGRESSION = "LogisticRegression"
    LINEAR_REGRESSION = "LinearRegression"
    DECISION_TREE = "DecisionTree"
    KNN = "KNN"
    STG = "STG"
    SVM = "SVM"
    TABNET = "TabNet"
    TABTRANSFORMER = "TabTransformer"
    DEEPFM = "DeepFM"
    VIME = "VIME"
    DANET = "DANet"
    NODE = "NODE"
    AUTOGLUON = "AutoGluon"


class ModelClass(StrEnum):
    BASE = 'base'
    GBDT = 'GBDT'
    NN = 'NN'
    ICLT = 'ICLT'


class DownstreamTask(StrEnum):
    ZEROSHOT = "zeroshot"
    FINETUNE = "finetune"



class BenchmarkName(StrEnum):
    DEBUG_CLASSIFICATION = "debug_classification"
    DEBUG_REGRESSION = "debug_regression"
    DEBUG_TABZILLA = "debug_tabzilla"

    CATEGORICAL_CLASSIFICATION = "categorical_classification"
    NUMERICAL_CLASSIFICATION = "numerical_classification"
    CATEGORICAL_REGRESSION = "categorical_regression"
    NUMERICAL_REGRESSION = "numerical_regression"
    CATEGORICAL_CLASSIFICATION_LARGE = "categorical_classification_large"
    NUMERICAL_CLASSIFICATION_LARGE = "numerical_classification_large"
    CATEGORICAL_REGRESSION_LARGE = "categorical_regression_large"
    NUMERICAL_REGRESSION_LARGE = "numerical_regression_large"

    TABZILLA_HARD = "tabzilla_hard"
    TABZILLA_HARD_MAX_TEN_CLASSES = "tabzilla_hard_max_ten_classes"
    TABZILLA_HAS_COMPLETED_RUNS = "tabzilla_has_completed_runs"


class BenchmarkOrigin(StrEnum):
    TABZILLA = "tabzilla"
    WHYTREES = "whytrees"


class GeneratorName(StrEnum):
    TABPFN = 'tabpfn'
    TREE = 'tree'
    RANDOMFOREST = 'randomforest'
    NEIGHBOR = 'neighbor'
    MIX = 'mix'
    PERLIN = 'perlin'
    MIX_7 = 'mix_7'
    MIX_6 = 'mix_6'
    MIX_5 = 'mix_5'
    MIX_5_GP = 'mix_5_gp'
    MIX_4 = 'mix_4'
    MIX_4_AG = 'mix_4_ag'
    LR = 'lr'
    POLY = 'poly'
    SAMPLE_RF = 'sample_rf'
    SAMPLE_GP = 'sample_gp'
    TABREPO = 'tabrepo'
    MIX_4_TABREPO = 'mix_4_tabrepo'
    MIX_4_TABPFNV2 = 'mix_4_tabpfnv2'


class MetricName(StrEnum):
    ACCURACY = "accuracy"
    F1 = "f1"
    AUC = "auc"
    MSE = "mse"
    MAE = "mae"
    R2 = "r2"
    LOG_LOSS = "log_loss"
    RMSE = "rmse"


class LossName(StrEnum):
    CROSS_ENTROPY = "cross_entropy"
    MSE = "mse"
    MAE = "mae"